from collections import Counter


punctuation_to_num = {'.': 27, ',': 28, '!': 29, "'": 30, "-": 31}
num_to_puncutation = {27: '.', 28: ',', 29: '!', 30:"'", 31: "-"}

def char_to_num(ch):
    # Check for space first.
    if ch == ' ':
        return 0
    
    # Convert character to lowercase for uniform processing.
    ch = ch.lower()
    
    # If it's a letter, map it from a=1, b=2, ..., z=26.
    if 'a' <= ch <= 'z':
        num = ord(ch) - ord('a') + 1  # a -> 1, b -> 2, ..., z -> 26
        return num
    
    # map punctuation to 5-bits as in punctuation_to_num
    elif ch in punctuation_to_num:
        return punctuation_to_num[ch]
    
    return ''

def num_to_char(num):
    if num == 0: #0 = space
        return ' '
    elif 1 <= num <= 26: #letters indexed by position in alphabet
        return chr(num + ord('a') + -1)
    elif num in num_to_puncutation: #map numbers to punctuation as in num_to_punctuation
        return num_to_puncutation[num]
    
#processes the string so that it only involves the allowed characters
def process(s):
    processed = ""
    s = s.lower() #convert to lowercase
    allowed_chars = {" "}
    allowed_chars.update(punctuation_to_num)  # Add all keys from punctuation_to_num.
    for ch in s:
        if 'a' <= ch <= 'z' or ch in allowed_chars:
            processed += ch
    return processed
    
def add_chars(c1:str, c2:str):
    assert len(c1) ==1
    assert len(c2) == 1
    s1 = char_to_num(c1)
    s2 = char_to_num(c2)
    return num_to_char((s1+s2)%32)

def shift_char_by_delta(delta: int, ch: str) -> str:
    num = char_to_num(ch)
    new_num = (num + delta) % 32
    return num_to_char(new_num)

def unshift_char_by_delta(delta: int, ch: str) -> str:
    num = char_to_num(ch)
    new_num = (num - delta) % 32
    return num_to_char(new_num)


def IC(s):
    assert s == process(s)
    char_counts = Counter(s)
    n = len(s)
    if n <= 1:
        print("string too short")
        return -1

    total_coincidences = 0
    for c in char_counts:
        total_coincidences += char_counts[c] * (char_counts[c]-1)
    return total_coincidences / (n*(n-1))



    
